VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsMessage"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' clsMessage
'
' 4/15/2001
' Kenrick Mock
' kenrick@acm.org
'
' This class exposes an interface to manage the contents of
' an individual message.  Currently only a number of terms
' from the message are stored.  These terms (actually, their IDs)
' are stored in a dictionary along with their frequency with
' respect to the local message (NOT global frequency).
'
' A comparison function is also here to compare one message to
' another and return a similarity value.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public dictTermIDFreq As Dictionary
Public dictTermIDPerc As Dictionary
Public sEntryID As String
Private Const NUMCOMPARETERMS = 30

Public Sub AddTerms(s As String)
    Dim l As Long
    Dim lIDs() As Long
    Dim lFreqs() As Long
    Dim sngPercents() As Single
    
    modFeatures.ExtractFeatures s, lIDs, lFreqs, sngPercents
    For l = 0 To UBound(lIDs)
        dictTermIDFreq(lIDs(l)) = lFreqs(l)
        dictTermIDPerc(lIDs(l)) = sngPercents(l)
    Next l
End Sub

Public Function GetIDs() As Variant
    GetIDs = dictTermIDFreq.Keys
End Function

Public Function GetFreqs() As Variant
    GetFreqs = dictTermIDFreq.Items
End Function

Public Function GetPercents() As Variant
    GetPercents = dictTermIDPerc.Items
End Function

' Open up the message with this particular Entry ID
Public Sub DisplayMessage()
    Dim oMsg As Outlook.MailItem
    If sEntryID <> "" Then
        Set oMsg = modGlobals.oNS.GetItemFromID(sEntryID)
        oMsg.Display
    End If
End Sub

' Gets the category from outlook for this item (actually all categories, but using
' this program, only one will be active).  Needs a valid EntryID
Public Function GetOutlookCategory() As String
    Dim oMsg As Outlook.MailItem
    If sEntryID <> "" Then
        Set oMsg = modGlobals.oNS.GetItemFromID(sEntryID)
        GetOutlookCategory = oMsg.Categories
    End If
End Function

' Sets the category in outlook for this item to the given string.
' Needs a valid EntryID
Public Sub SetOutlookCategory(sCat As String)
    Dim oMsg As Outlook.MailItem
    If sEntryID <> "" Then
        Set oMsg = modGlobals.oNS.GetItemFromID(sEntryID)
        oMsg.Categories = sCat
        oMsg.Save
    End If
End Sub

' Compare with respect to a destination message, using the
' percentages from the destination message.
' Currently only compares the top 30 terms, selected arbitrarily...
Public Function CompareToMessage(targMsg As clsMessage) As Single
    Dim vQueryTerms As Variant
    
    vQueryTerms = dictTermIDFreq.Keys
    If UBound(vQueryTerms) > NUMCOMPARETERMS Then
        ReDim Preserve vQueryTerms(NUMCOMPARETERMS)
    End If
    CompareToMessage = modFeatures.QueryRelevanceByDict(vQueryTerms, targMsg.dictTermIDPerc)
End Function

Private Sub Class_Initialize()
    Set dictTermIDFreq = New Dictionary
    Set dictTermIDPerc = New Dictionary
    sEntryID = ""
End Sub

Private Sub Class_Terminate()
    Set dictTermIDPerc = Nothing
    Set dictermidfreq = Nothing
End Sub
