Attribute VB_Name = "modFeatures"
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' modFeatures
'
' 4/15/2001
' Kenrick Mock
' kenrick@acm.org
'
' This module contains feature information and classification
' routines.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Option Explicit
Option Compare Text
Public strHash As New STRHASHLib.CStrHash

' Turn input into array of tokens
' Simple version for now, assumes input is all text and not html
' ***** Hack -- also removes words <= 3 letters long here!
' ***** we should do this in the stoplist routine but by then it's already a numeric ID...
Public Sub Tokenize(sIn As String, sArrayOut() As String)
    Dim sSplitAry() As String
    Dim l As Long
    Dim lOut As Long
    
    ReDim sArrayOut(0)
    sIn = Trim(sIn)
    If sIn = "" Then
        Exit Sub
    End If
    
    sSplitAry = Split(sIn, " ")
    ReDim sArrayOut(UBound(sSplitAry))
    lOut = 0
    For l = 0 To UBound(sSplitAry)
        If Len(sSplitAry(l)) > 3 Then
            If Len(sSplitAry(l)) > 20 Then
                sArrayOut(lOut) = sSplitAry(l)
            Else
                sArrayOut(lOut) = Left(sSplitAry(l), 20)
            End If
            lOut = lOut + 1
        End If
    Next l
    If lOut > 0 Then
        ReDim Preserve sArrayOut(lOut - 1)
    End If
End Sub

'Strip ALL punctuation out of the string
'and convert to lowercase.
'We will probalby want a different one when we're being smarter
'and looking for periods and that kind of thing
Public Sub StripAllPunctuationToLower(ByRef sInOut As String)
    Dim sNewString As String
    Dim l As Long
    
    For l = 1 To Len(sInOut)
        If Asc(Mid(sInOut, l, 1)) >= Asc("a") And Asc(Mid(sInOut, l, 1)) <= Asc("z") Then
           ' Save, its lowercase letters
           sNewString = sNewString & Mid(sInOut, l, 1)
        ElseIf Asc(Mid(sInOut, l, 1)) >= Asc("A") And Asc(Mid(sInOut, l, 1)) <= Asc("Z") Then
           ' Save, to lowercase letters
           sNewString = sNewString & Chr(Asc(Mid(sInOut, l, 1)) + 32)
        Else
           ' Save a space instead for all punctuation
           sNewString = sNewString & " "
        End If
    Next l
    sInOut = sNewString
End Sub

'Tokenize, remove stop words, count freqencies, return them
Public Sub ExtractFeatures(sIn As String, ByRef lTermIDAry() As Long, ByRef lTermFreqs() As Long, ByRef sTermPercents() As Single)
    Dim sFeatAry() As String
    Dim lStopListedAry() As Long
    Dim dictFeatureCount As Dictionary
    Dim dictFeatures As Dictionary
    Dim vTerms As Variant
    Dim vCounts As Variant
    Dim lCount As Long
    Dim lTotal As Long
    Dim l As Long
    Dim sngFreq As Single
    Dim vTermPercents As Variant
    Dim vTermAry As Variant
    Dim vStrings As Variant
    Dim vIDs As Variant
    
    ReDim lTermIDAry(0)
    ReDim sTermPercents(0)
    ReDim lTermFreqs(0)
    ' Tokenize
    StripAllPunctuationToLower sIn
    Tokenize sIn, sFeatAry
    ' Convert to IDs
    vStrings = sFeatAry
    strHash.HashStringToID vStrings, vIDs
    ' Stoplist
    ReDim lTermIDAry(UBound(vIDs))
    For l = 0 To UBound(vIDs)
        lTermIDAry(l) = CLng(vIDs(l))
    Next l
    modStopList.RemoveStopWordIDs lTermIDAry, lStopListedAry
    ' Quit if we have nothing
    If UBound(lStopListedAry) = 0 Then
        Exit Sub
    End If
    
    ' Count up frequencies for each and put it in a dictionary
    lTotal = 0
    Set dictFeatureCount = New Dictionary
    For l = 1 To UBound(lStopListedAry)
        lTotal = lTotal + 1
        If dictFeatureCount.Exists(lStopListedAry(l)) Then
            ' Increment it
            lCount = dictFeatureCount(lStopListedAry(l))
            lCount = lCount + 1
            dictFeatureCount(lStopListedAry(l)) = lCount
        Else
            ' Add it
            dictFeatureCount.Add lStopListedAry(l), 1
        End If
    Next l
    
    ' Go through each item and store a percentage in the feature dictionary
    Set dictFeatures = New Dictionary
    vTerms = dictFeatureCount.Keys
    vCounts = dictFeatureCount.Items
    For l = 0 To UBound(vTerms)
        sngFreq = CSng(vCounts(l) / lTotal)
        dictFeatures.Add vTerms(l), sngFreq
    Next l
    
    ' Sort and return arrays
    vTermAry = dictFeatures.Keys
    vTermPercents = dictFeatures.Items
    modSorters.sortFeatures vTermAry, vTermPercents, 0, UBound(vTermPercents)
    ' Copy to outputs
    ReDim lTermIDAry(UBound(vTermAry))
    ReDim sTermPercents(UBound(vTermAry))
    ReDim lTermFreqs(UBound(vTermAry))
    For l = 0 To UBound(vTermAry)
        lTermIDAry(l) = CLng(vTermAry(l))
        sTermPercents(l) = CSng(vTermPercents(l))
        lTermFreqs(l) = CLng(dictFeatureCount(vTermAry(l)))
    Next l
    
    Set dictFeatures = Nothing
    Set dictFeatureCount = Nothing
End Sub

' Compute relevance between a set of query terms IDs and input document
' Will be a small number
'
' Was using the cosine coefficient, trying a simpler technique for now...
' assymetric comparison, only use weights from the query.
' Simplifies our data collection a bit, but a weaker comparison metric.
Public Function QueryRelevance(ByRef lQueryTerms() As Long, ByRef lTermAry() As Long, ByRef sTermPercents() As Single) As Single
    Dim sngRel As Single
    Dim l As Long
    Dim dictTerms As Dictionary
    
    sngRel = 0
    Set dictTerms = New Dictionary
    ' Turn terms into dictionary
    For l = 0 To UBound(lTermAry)
        dictTerms.Add lTermAry(l), sTermPercents(l)
    Next l
    ' Just sum up the values for those that match in the query
    For l = 0 To UBound(lQueryTerms)
        If dictTerms.Exists(lQueryTerms(l)) Then
            sngRel = sngRel + dictTerms(lQueryTerms(l))
        End If
    Next l
    Set dictTerms = Nothing
    QueryRelevance = sngRel
End Function

' Same as above but passing in a dictionary
Public Function QueryRelevanceByDict(ByRef lQueryTerms As Variant, ByRef dictSrcIDsPerc As Dictionary) As Single
    Dim sngRel As Single
    Dim l As Long
    
    sngRel = 0
    ' Just sum up the values for those that match in the query
    For l = 0 To UBound(lQueryTerms)
        If dictSrcIDsPerc.Exists(lQueryTerms(l)) Then
            sngRel = sngRel + dictSrcIDsPerc(lQueryTerms(l))
        End If
    Next l
    QueryRelevanceByDict = sngRel
End Function

' Strip last word.  Used when truncating a string to N characters,
' if we chopped a word in half, move back to the previous space
Public Function StripLastWord(s As String) As String
    Dim l As Long
    For l = Len(s) To 1 Step -1
        If Mid(s, l, 1) = " " Then
            StripLastWord = Left(s, l - 1)
            Exit Function
        End If
    Next l
    StripLastWord = s
End Function

'**************** Outlook-specific Classification routines follow ****************

' Given a folder, go through all the messages, collecting text information about each
' one and build up our nearest neighbor classifier
Public Sub CollectFolderText(ByRef curFolder As Outlook.MAPIFolder, ByRef myFolder As clsFolder)
    Dim l As Long
    Dim oMsg As Outlook.MailItem
    Dim sSender As String
    Dim sSubj As String
    Dim sBody As String
    Dim sStatus As String
    
    For l = 1 To curFolder.Items.Count
        sStatus = "Item " & Str(l) & " of " & Str(curFolder.Items.Count) & " : " & curFolder.Name
        frmLoading.UpdateFolderName sStatus
        If l Mod 10 = 0 Then
            frmLoading.TogglePicture
        End If
        
        Set oMsg = curFolder.Items.Item(l)
        sSender = oMsg.SenderName
        sSubj = oMsg.Subject
        sBody = Left(oMsg.Body, MAXBODYCHARS)
        sBody = modFeatures.StripLastWord(sBody)
        myFolder.AddMessage oMsg.EntryID, sSender, sSubj, sBody
        DoEvents
    Next l
End Sub

  
' Recursively enumerate through the folders, collecting the folder
' name and the messages contained in each folder
Public Sub CollectFolderNames(curFolder As Outlook.MAPIFolder)
    Dim l As Long
    Dim clsFol As clsFolder
    
    For l = 1 To curFolder.Folders.Count
        Set clsFol = New clsFolder
        clsFol.sFolderName = curFolder.Folders.Item(l).Name
        modFeatures.CollectFolderText curFolder.Folders.Item(l), clsFol
        modGlobals.dictFolders.Add curFolder.Folders.Item(l).Name, clsFol
        modFeatures.CollectFolderNames curFolder.Folders.Item(l)
        DoEvents
    Next l
End Sub

' Enumerate through messages in the inbox (non-recursive)
Public Sub CollectInboxInfo(curFolder As Outlook.MAPIFolder)
    Dim clsFol As clsFolder
        
    Set clsFol = New clsFolder
    clsFol.sFolderName = curFolder.Name
    modFeatures.CollectFolderText curFolder, clsFol
    modGlobals.dictFolders.Add curFolder.Name, clsFol
End Sub

' Given a message classify it in one of the folders
' assuming it doesn't have a folder category already
Public Sub ClassifyIntoFolder(curMsg As clsMessage)
    Dim f As clsFolder
    Dim fBestFolder As clsFolder
    Dim l As Long
    Dim dMax As Double
    Dim dRel As Double
    
    ' Quit immediately if we find an existing outlook category
    If Trim(curMsg.GetOutlookCategory) <> "" Then
        Exit Sub
    End If
    
    ' Find best folder
    dMax = -1
    For l = 0 To modGlobals.dictFolders.Count - 1
        Set f = modGlobals.dictFolders.Items(l)
        ' Continue only if this isn't the inbox folder (can't classify into the inbox!)
        If (f.sFolderName <> modGlobals.objInbox.Name) Then
            dRel = f.CompareMessage(curMsg)
            If dRel > dMax Then
                dMax = dRel
                Set fBestFolder = f
            End If
        End If
        DoEvents
    Next l
    
    ' Classify it
    curMsg.SetOutlookCategory fBestFolder.sFolderName
    
    Set fBestFolder = Nothing
    Set f = Nothing
End Sub


' Try classifying everything in the inbox
Public Sub ClassifyInboxMsgs(Optional bReclassify As Boolean = True)
  Dim oItem As Object
  Dim oMsg As Outlook.MailItem
  Dim cMsg As clsMessage
  Dim l As Long
  Dim sSender As String
  Dim sSubj As String
  Dim sBody As String
    
  On Error Resume Next      ' Message might be moved while we're processing
  For l = 1 To modGlobals.objInbox.Items.Count
        Set oMsg = objInbox.Items.Item(l)
        ' Skip unless we find an existing outlook category
        If Trim(oMsg.Categories) = "" Then
            Set cMsg = New clsMessage
            sSender = oMsg.SenderName
            sSubj = oMsg.Subject
            sBody = Left(oMsg.Body, MAXBODYCHARS)
            sBody = modFeatures.StripLastWord(sBody)
            cMsg.sEntryID = oMsg.EntryID
            cMsg.AddTerms sSender + " " + sSubj + " " + sBody
            modFeatures.ClassifyIntoFolder cMsg
            Set cMsg = Nothing
        End If
        DoEvents
  Next l
  ' KLUDGE ALERT!
  '
  ' Set up a timer to re-check inbox msgs again in a minute.
  ' Unfortunately there does not appear to be an event for
  ' when mail delivery is FINISHED.   When multiple email is
  ' received at once, there may be a race condition where this
  ' classification routine finishes before the rest of the email
  ' is delivered, resulting in classification of only one or two emails.
  '
  ' Of course this is not guaranteed to catch all new emails either,
  ' but once a new message is received it should catch up.
  If bReclassify = True Then
    frmLoading.TimerRescanInbox.Enabled = True
    frmLoading.TimerRescanInbox.Interval = 30 * 1000
  End If
  On Error GoTo 0
End Sub

