// CStrHash.cpp : Implementation of CCStrHash
//
// This project builds a COM object that exposes a single method:  HashStringToID.
// This hashes a string into a 32 bit value using the mix algorithm below.
// This has been implemented as a C++ COM object for the purpose of
// providing a string->long hash for use from Visual Basic.
//
// As of this writing, this was needed since VB 6.0 lacks the necessary
// primitives to logically AND, OR, XOR individual bits.  (I understand this has
// changed in VB.NET).  As a result, this object may not be needed in the future
// if the hashing algorithm below is desired.  
//
// It is a fairly fast hashing algorithm that produces few collisions on text data.
// Feel free to modify/use this code as you see fit.
//
// Kenrick Mock
// kenrick@acm.org
// 3/15/2001
//

#include "stdafx.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "STLVarArrayUtils.h"
#include "StrHash.h"
#include "CStrHash.h"
#include "hash.h"


/////////////////////////////////////////////////////////////////////////////
// CCStrHash


HRESULT CCStrHash::HashStringToID(VARIANT vStrArray, VARIANT *pvIDArray)
{	
	vector<string> strAry;
	vector<string>::iterator strIter;
	vector<long> lAry;
	char s[1000];
	long lID;

	VariantClear(pvIDArray);
	CSTLVarArrayUtils::VarArrayToVector(strAry, vStrArray);
	for (strIter = strAry.begin(); strIter!=strAry.end(); strIter++) {
		strncpy(s, strIter->c_str(), 999);
		s[999]='\0';
		lID = (long) myhash((unsigned char *) s, strlen(s), 58913);
		lAry.push_back(lID);
	}
	CSTLVarArrayUtils::VectorToVarArray(lAry, pvIDArray);
	return S_OK;
}


#define hashsize(n) ((ub4)1<<(n))
#define hashmask(n) (hashsize(n)-1)

/*
--------------------------------------------------------------------
mix -- mix 3 32-bit values reversibly.
For every delta with one or two bits set, and the deltas of all three
  high bits or all three low bits, whether the original value of a,b,c
  is almost all zero or is uniformly distributed,
* If mix() is run forward or backward, at least 32 bits in a,b,c
  have at least 1/4 probability of changing.
* If mix() is run forward, every bit of c will change between 1/3 and
  2/3 of the time.  (Well, 22/100 and 78/100 for some 2-bit deltas.)
mix() takes 36 machine instructions, but only 18 cycles on a superscalar
  machine (like a Pentium or a Sparc).  No faster mixer seems to work,
  that's the result of my brute-force search.  There were about 2^^68
  hashes to choose from.  I only tested about a billion of those.
--------------------------------------------------------------------
*/
#define mix(a,b,c) \
{ \
  a -= b; a -= c; a ^= (c>>13); \
  b -= c; b -= a; b ^= (a<<8); \
  c -= a; c -= b; c ^= (b>>13); \
  a -= b; a -= c; a ^= (c>>12);  \
  b -= c; b -= a; b ^= (a<<16); \
  c -= a; c -= b; c ^= (b>>5); \
  a -= b; a -= c; a ^= (c>>3);  \
  b -= c; b -= a; b ^= (a<<10); \
  c -= a; c -= b; c ^= (b>>15); \
}

/*
--------------------------------------------------------------------
hash() -- hash a variable-length key into a 32-bit value
  k       : the key (the unaligned variable-length array of bytes)
  len     : the length of the key, counting by bytes
  initval : can be any 4-byte value
Returns a 32-bit value.  Every bit of the key affects every bit of
the return value.  Every 1-bit and 2-bit delta achieves avalanche.
About 6*len+35 instructions.

The best hash table sizes are powers of 2.  There is no need to do
mod a prime (mod is sooo slow!).  If you need less than 32 bits,
use a bitmask.  For example, if you need only 10 bits, do
  h = (h & hashmask(10));
In which case, the hash table should have hashsize(10) elements.

If you are hashing n strings (ub1 **)k, do it like this:
  for (i=0, h=0; i<n; ++i) h = hash( k[i], len[i], h);

By Bob Jenkins, 1996.  bob_jenkins@burtleburtle.net.  You may use this
code any way you wish, private, educational, or commercial.  It's free.

See http://burtleburtle.net/bob/hash/evahash.html
Use for hash table lookup, or anything where one collision in 2^^32 is
acceptable.  Do NOT use for cryptographic purposes.
--------------------------------------------------------------------
*/

ub4 CCStrHash::myhash( register ub1 *k, register ub4 length, register ub4 initval)
{
   register ub4 a,b,c,len;

   /* Set up the internal state */
   len = length;
   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
   c = initval;         /* the previous hash value */

   /*---------------------------------------- handle most of the key */
   while (len >= 12)
   {
      a += (k[0] +((ub4)k[1]<<8) +((ub4)k[2]<<16) +((ub4)k[3]<<24));
      b += (k[4] +((ub4)k[5]<<8) +((ub4)k[6]<<16) +((ub4)k[7]<<24));
      c += (k[8] +((ub4)k[9]<<8) +((ub4)k[10]<<16)+((ub4)k[11]<<24));
      mix(a,b,c);
      k += 12; len -= 12;
   }

   /*------------------------------------- handle the last 11 bytes */
   c += length;
   switch(len)              /* all the case statements fall through */
   {
   case 11: c+=((ub4)k[10]<<24);
   case 10: c+=((ub4)k[9]<<16);
   case 9 : c+=((ub4)k[8]<<8);
      /* the first byte of c is reserved for the length */
   case 8 : b+=((ub4)k[7]<<24);
   case 7 : b+=((ub4)k[6]<<16);
   case 6 : b+=((ub4)k[5]<<8);
   case 5 : b+=k[4];
   case 4 : a+=((ub4)k[3]<<24);
   case 3 : a+=((ub4)k[2]<<16);
   case 2 : a+=((ub4)k[1]<<8);
   case 1 : a+=k[0];
     /* case 0: nothing left to add */
   }
   mix(a,b,c);
   /*-------------------------------------------- report the result */
   return c;
}


void main()
{
}

/*
main()
{
 long l,lInit;
 char s[500];
 FILE *f;

 srand(5123);
 lInit = (long) rand();

 f=fopen("/usr/dict/words","r");
 while (fgets(s,500,f)!=NULL) {
	s[strlen(s)-1]='\0';
 	l=(long) myhash(s, strlen(s), lInit);
	if (l<0) l=l*-1;
 	printf("%ld\n",l);
 }
 fclose(f);
}


On this data, there were two duplicates.
The data took up 748893K.  There were 87311 words.
The numeric representation only uses 349244K, about
half the storage.  Big bonus is much faster processing
than string functions.

On 200Mhz Pentium pro:
   strcmp on 500,000 three-letter strings took 0.19uSec on ave
   longcmp on 500,000 numbers took 0.02usec on ave


*/
